/* GKrellM
|  Copyright (C) 1999-2002 Bill Wilson
|
|  Author:  Bill Wilson    bill@gkrellm.net
|  Latest versions might be found at:  http://gkrellm.net
|
|  This program is free software which I release under the GNU General Public
|  License. You may redistribute and/or modify this program under the terms
|  of that license as published by the Free Software Foundation; either
|  version 2 of the License, or (at your option) any later version.
|
|  This program is distributed in the hope that it will be useful,
|  but WITHOUT ANY WARRANTY; without even the implied warranty of
|  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|  GNU General Public License for more details. Version 2 is in the
|  COPYRIGHT file in the top level directory of this distribution.
| 
|  To get a copy of the GNU General Puplic License, write to the Free Software
|  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gkrellm.h"
#include "gkrellm_private_proto.h"
#include "gkrellm_threads.h"

static Monitor	*mon_clock,
				*mon_cal;

static Panel	*pclock,
				*pcal;

static Decal	*d_wday,
				*d_mday,
				*d_month,
				*d_clock,
				*d_seconds;


static Launcher	clock_launch,
				cal_launch;

typedef struct
	{
	TextStyle	ts;
	gint		lbearing,
				rbearing,
				width,
				ascent,
				descent;
	}
	Extents;

static gboolean		cal_enable,
					clock_enable,
					clock_12hr_sec_enable,
					clock_24hr_sec_enable,
					clock_24hr_mode,
                    loop_chime_enable;

static gint			x_clock_seconds,
					x_clock_no_seconds;


static gchar       *hour_chime_command,
                   *quarter_chime_command;

typedef struct
    {
    gchar *command;
    gint   count;
    }
    ChimeData;

static Extents		wday_extents,
					mday_extents,
					month_extents,
					time_extents,
					sec_extents,
					ampm_extents;

static gint			clock_style_id,
					cal_style_id;

struct tm			gkrellm_current_tm;

struct tm *
gkrellm_get_current_time(void)
	{
	return &gkrellm_current_tm;
	}

static void
string_extents(gchar *string, Extents *ext)
	{
	gdk_string_extents(ext->ts.font, string, &ext->lbearing, &ext->rbearing,
			&ext->width, &ext->ascent, &ext->descent);
	}

static void
chime_func(gpointer data)
{
    ChimeData *chime = (ChimeData *)data;
    int counter;

    if (strlen(chime -> command)) 
        {
        if (chime -> count > 12) 
            chime -> count -= 12;

        for (counter = 0; counter < chime -> count; counter ++)
            gkrellm_system(chime -> command);
        }
    g_free(chime -> command);
    g_free(chime);
}

static void
draw_cal(struct tm *t)
	{
	gchar	buf[32];
	gint	w;

	if (!cal_enable)
		return;
	strftime(buf, sizeof(buf), "%a", t);  /* Abbreviated weekday name */
	w = gdk_string_width(d_wday->text_style.font, buf);
	d_wday->x_off = (w < d_wday->w) ? (d_wday->w - w) / 2 : 0;
	gkrellm_draw_decal_text(pcal, d_wday, buf, -1);

	/*strftime(buf, sizeof(buf), "%e", t);*/  /* Day in month 1-31 */
	snprintf(buf, sizeof(buf), "%d", t->tm_mday);
	w = gdk_string_width(d_mday->text_style.font, buf);
	d_mday->x_off = (w < d_mday->w) ? (d_mday->w - w) / 2 : 0;
	gkrellm_draw_decal_text(pcal, d_mday, buf, -1);

	strftime(buf, sizeof(buf), "%b", t); /* Abbreviated month name */
	gkrellm_draw_decal_text(pcal, d_month, buf, -1);
	gkrellm_draw_panel_layers(pcal);
	}

static void
draw_clock(struct tm *t, gboolean update_minutes)
	{
	gchar	buf[32], sec_ampm[32];
	gint	w;

	if (!clock_enable)
		return;
	if (update_minutes)
		{
		if (clock_24hr_mode)
			strftime(buf, sizeof(buf), "%k:%M", t);   /* 24 hour:minute */
		else
			strftime(buf, sizeof(buf), "%l:%M", t);   /* 12 hour:minute */
		w = gdk_string_width(d_clock->text_style.font, buf);
		d_clock->x_off = (w < d_clock->w) ? (d_clock->w - w) / 2 : 0;
		gkrellm_draw_decal_text(pclock, d_clock, buf, -1);
		}
	if (   (!clock_24hr_mode && clock_12hr_sec_enable)
		|| (clock_24hr_mode && clock_24hr_sec_enable)
	   )
		{
		strftime(sec_ampm, sizeof(sec_ampm), "%S", t);	/* seconds 00-59 */
		gkrellm_draw_decal_text(pclock, d_seconds, sec_ampm, -1);
		}
	else if (! clock_24hr_mode)
		{
		strftime(sec_ampm, sizeof(sec_ampm), "%p", t);	/* AM/PM */
		gkrellm_draw_decal_text(pclock, d_seconds, sec_ampm, -1);
		}
	gkrellm_draw_panel_layers(pclock);
	}

static void
update_clock(void)
	{
	gboolean	update_minutes = FALSE;
	static gint	min_prev, hour_prev = -1;
    ChimeData  *chime;

	if (!GK.second_tick)
		return;
	if (   gkrellm_current_tm.tm_min != min_prev
		|| gkrellm_current_tm.tm_hour != hour_prev
	   )
		{
		update_minutes = TRUE;
		draw_cal(&gkrellm_current_tm);
        if (   gkrellm_current_tm.tm_hour != hour_prev
            && hour_prev != -1
           ) 
            {
                if (hour_chime_command && *hour_chime_command) 
                    {
                    chime = g_new0(ChimeData, 1);
                    chime -> command = strdup(hour_chime_command);
                    chime -> count = loop_chime_enable ? gkrellm_current_tm.tm_hour : 1;
                    new_thread(chime_func, chime);
                    }
            }
        else
            {
                if (   (gkrellm_current_tm.tm_min % 15) == 0
                    && quarter_chime_command && *quarter_chime_command
                   ) 
                    {
                    
                    chime = g_new0(ChimeData, 1);
                    chime -> command = strdup(quarter_chime_command);
                    chime -> count = 1;
                    new_thread(chime_func, chime);
                    }
            }        
		}
	draw_clock(&gkrellm_current_tm, update_minutes);
	min_prev = gkrellm_current_tm.tm_min;
	hour_prev = gkrellm_current_tm.tm_hour;
	}

static gint
expose_event(GtkWidget *widget, GdkEventExpose *ev)
	{
	GdkPixmap	*pixmap = NULL;

	if (widget == pcal->drawing_area)
		pixmap = pcal->pixmap;
	else if (widget == pclock->drawing_area)
		pixmap = pclock->pixmap;
	if (pixmap)
		gdk_draw_pixmap(widget->window, GK.draw1_GC, pixmap,
				ev->area.x, ev->area.y, ev->area.x, ev->area.y,
				ev->area.width, ev->area.height);
	return FALSE;
	}

static gint
cb_panel_press(GtkWidget *widget, GdkEventButton *ev)
	{
	if (ev->button == 3)
		gkrellm_open_config_window(mon_clock);
	return TRUE;
	}

static void
create_calendar(GtkWidget *vbox, gint first_create)
	{
	Style	*style;
	gint	w;

	if (first_create)
		pcal = gkrellm_panel_new0();
	style = gkrellm_meter_style(cal_style_id);

	wday_extents.ts = *gkrellm_meter_textstyle(cal_style_id);
	string_extents("Wed", &wday_extents);
	mday_extents.ts = *gkrellm_meter_alt_textstyle(cal_style_id);
	string_extents("28", &mday_extents);
	month_extents.ts = *gkrellm_meter_textstyle(cal_style_id);
	string_extents("Aug", &month_extents);

	d_wday = gkrellm_create_decal_text(pcal, "88", &wday_extents.ts,
				style, 0, -1 /*top margin*/, wday_extents.width + 2);
	d_mday = gkrellm_create_decal_text(pcal, "88", &mday_extents.ts,
				style, 0, -1 /*top margin*/, mday_extents.width + 2);
	d_month = gkrellm_create_decal_text(pcal, "88", &month_extents.ts,
				style, 0, -1 /*top margin*/, month_extents.width + 2);
	w = d_wday->w + d_mday->w + d_month->w;
	d_wday->x = (gkrellm_chart_width() - w) / 2;
	d_mday->x = d_wday->x + d_wday->w;
	d_month->x = d_mday->x + d_mday->w;
	if (d_wday->h < d_mday->h)
		{
		d_wday->y = d_mday->y + d_mday->h - d_wday->h - 1;
		d_month->y = d_wday->y;
		}
	else if (d_mday->h < d_wday->h)
		d_mday->y = d_wday->y + d_wday->h - d_mday->h - 1;
	gkrellm_panel_configure(pcal, NULL, style);
	gkrellm_panel_create(vbox, mon_cal, pcal);

	if (first_create)
		{
		/* Help the motion out hack. If starting a move in host panel and mouse
		|  jerks into the pclock/pcal drawing areas, we stop moving unless this:
		*/
		extern void gkrellm_motion(GtkWidget *, GdkEventMotion *, gpointer);

		gtk_signal_connect(GTK_OBJECT(pcal->drawing_area),
				"motion_notify_event", (GtkSignalFunc) gkrellm_motion, NULL);

		gtk_signal_connect(GTK_OBJECT (pcal->drawing_area), "expose_event",
				(GtkSignalFunc) expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT(pcal->drawing_area),
				"button_press_event", (GtkSignalFunc) cb_panel_press, NULL);
		}
	gkrellm_setup_launcher(pcal, &cal_launch, METER_PANEL_TYPE, 0);

	if (!cal_enable)
		gkrellm_panel_hide(pcal);

	draw_cal(&gkrellm_current_tm);
	}

static void
set_seconds_display()
	{
	if (   !clock_24hr_mode
		|| (clock_24hr_mode && clock_24hr_sec_enable)
	   )
		{
		gkrellm_move_decal(pclock, d_clock, x_clock_seconds, d_clock->y);
		gkrellm_make_decal_visible(pclock, d_seconds);
		}
	else
		{
		gkrellm_move_decal(pclock, d_clock, x_clock_no_seconds, d_clock->y);
		gkrellm_make_decal_invisible(pclock, d_seconds);
		}
	}

static void
create_clock(GtkWidget *vbox, gint first_create)
	{
	Style	*style;

	if (first_create)
		pclock = gkrellm_panel_new0();
	style = gkrellm_meter_style(clock_style_id);

	time_extents.ts = *gkrellm_meter_textstyle(clock_style_id);
	string_extents("00:00", &time_extents);
	sec_extents.ts = *gkrellm_meter_alt_textstyle(clock_style_id);
	string_extents("8M", &sec_extents);
	ampm_extents.ts = *gkrellm_meter_alt_textstyle(clock_style_id);
	string_extents("8M", &ampm_extents);

	d_clock = gkrellm_create_decal_text(pclock, "88", &time_extents.ts,
				style, 0, -1 /*top margin*/, time_extents.width + 2);
	d_seconds = gkrellm_create_decal_text(pclock, "88", &sec_extents.ts,
				style, 0, -1, sec_extents.width + 2);

	x_clock_seconds = (gkrellm_chart_width() - d_clock->w - d_seconds->w) / 2;
	x_clock_no_seconds = (gkrellm_chart_width() - d_clock->w) / 2;
	d_seconds->x = x_clock_seconds + d_clock->w + 2;
	d_seconds->y = d_clock->y + d_clock->h - d_seconds->h;

	gkrellm_panel_configure(pclock, NULL, style);
	gkrellm_panel_create(vbox, mon_clock, pclock);

	if (first_create)
		{
		gtk_signal_connect(GTK_OBJECT (pclock->drawing_area), "expose_event",
				(GtkSignalFunc) expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT(pclock->drawing_area),
				"button_press_event", (GtkSignalFunc) cb_panel_press, NULL);
		}
	gkrellm_setup_launcher(pclock, &clock_launch, METER_PANEL_TYPE, 0);

	set_seconds_display();
	if (!clock_enable)
		{
		gkrellm_panel_hide(pclock);
		hide_spacers(mon_clock);
		}
	draw_clock(&gkrellm_current_tm, TRUE);
	}

#define	CLOCK_CONFIG_KEYWORD	"clock_cal"

static void
save_clock_cal_config(FILE *f)
	{
	fprintf(f, "%s clock_launch %s\n", CLOCK_CONFIG_KEYWORD,
				clock_launch.command);
	fprintf(f, "%s clock_tooltip %s\n", CLOCK_CONFIG_KEYWORD,
				clock_launch.tooltip_comment);
	fprintf(f, "%s cal_launch %s\n", CLOCK_CONFIG_KEYWORD,
				cal_launch.command);
	fprintf(f, "%s cal_tooltip %s\n", CLOCK_CONFIG_KEYWORD,
				cal_launch.tooltip_comment);
	fprintf(f, "%s hour_chime_command %s\n", CLOCK_CONFIG_KEYWORD,
				hour_chime_command);
	fprintf(f, "%s quarter_chime_command %s\n", CLOCK_CONFIG_KEYWORD,
				quarter_chime_command);
    fprintf(f, "%s loop_chime_enable %d\n", CLOCK_CONFIG_KEYWORD,
               loop_chime_enable);
	fprintf(f, "%s clock_options %d %d %d %d\n", CLOCK_CONFIG_KEYWORD,
				clock_enable, clock_12hr_sec_enable, clock_24hr_mode, 
				clock_24hr_sec_enable);
	fprintf(f, "%s cal_options %d\n", CLOCK_CONFIG_KEYWORD,
				cal_enable);
	}

static void
load_clock_cal_config(gchar *arg)
	{
	gchar	config[32], item[CFG_BUFSIZE];
	gint	n;

	n = sscanf(arg, "%31s %[^\n]", config, item);
	if (n == 2)
		{
		if (!strcmp(config, "clock_launch"))
			clock_launch.command = g_strdup(item);
		else if (!strcmp(config, "clock_tooltip"))
			clock_launch.tooltip_comment = g_strdup(item);
		else if (!strcmp(config, "cal_launch"))
			cal_launch.command = g_strdup(item);
		else if (!strcmp(config, "cal_tooltip"))
			cal_launch.tooltip_comment = g_strdup(item);
		else if (!strcmp(config, "hour_chime_command"))
			gkrellm_dup_string(&hour_chime_command, item);
		else if (!strcmp(config, "quarter_chime_command"))
			gkrellm_dup_string(&quarter_chime_command, item);
		else if (!strcmp(config, "loop_chime_enable"))
			sscanf(item, "%d", &loop_chime_enable);
		else if (!strcmp(config, "clock_options"))
			sscanf(item, "%d %d %d %d", &clock_enable, &clock_12hr_sec_enable,
					&clock_24hr_mode, &clock_24hr_sec_enable);
		else if (!strcmp(config, "cal_options"))
			sscanf(item, "%d", &cal_enable);
		}
	}

/* --------------------------------------------------------------------- */
static GtkWidget	*cal_launch_entry,
					*cal_tooltip_entry,
					*clock_launch_entry,
					*clock_tooltip_entry,
                    *hour_chime_entry,
                    *quarter_chime_entry,
					*clock_enable_button,
					*cal_enable_button,
					*loop_chime_button,
					*clock_24hr_button,
					*clock_24hr_sec_button,
					*clock_12hr_sec_button;


static void
apply_clock_cal_config()
	{
	clock_24hr_mode = GTK_TOGGLE_BUTTON (clock_24hr_button)->active;
	clock_12hr_sec_enable = GTK_TOGGLE_BUTTON (clock_12hr_sec_button)->active;
	clock_24hr_sec_enable = GTK_TOGGLE_BUTTON (clock_24hr_sec_button)->active;

	clock_enable = GTK_TOGGLE_BUTTON (clock_enable_button)->active;
	if (clock_enable)
		{
		gkrellm_panel_show(pclock);
		show_spacers(mon_clock);
		}
	else
		{
		gkrellm_panel_hide(pclock);
		hide_spacers(mon_clock);
		}

	cal_enable = GTK_TOGGLE_BUTTON (cal_enable_button)->active;
	if (cal_enable)
		gkrellm_panel_show(pcal);
	else
		gkrellm_panel_hide(pcal);

    gkrellm_dup_string(&hour_chime_command,
			gkrellm_entry_get_text(&hour_chime_entry));
    gkrellm_dup_string(&quarter_chime_command,
			gkrellm_entry_get_text(&quarter_chime_entry));
	loop_chime_enable = GTK_TOGGLE_BUTTON(loop_chime_button)->active;

	gkrellm_apply_launcher(&cal_launch_entry, &cal_tooltip_entry, pcal,
					&cal_launch, gkrellm_launch_button_cb);
	gkrellm_apply_launcher(&clock_launch_entry, &clock_tooltip_entry, pclock,
					&clock_launch, gkrellm_launch_button_cb);

	set_seconds_display();
	draw_cal(&gkrellm_current_tm);
	draw_clock(&gkrellm_current_tm, TRUE);
	}

static void
set_seconds_buttons_sensitivity(gboolean mode_24hr)
	{
	if (mode_24hr)
		{
		gtk_widget_set_sensitive(clock_24hr_sec_button, TRUE);
		gtk_widget_set_sensitive(clock_12hr_sec_button, FALSE);
		}
	else
		{
		gtk_widget_set_sensitive(clock_24hr_sec_button, FALSE);
		gtk_widget_set_sensitive(clock_12hr_sec_button, TRUE);
		}
	}

static void cb_clock_24hr(GtkToggleButton *tb, gpointer user_data)
	{
	set_seconds_buttons_sensitivity(tb->active);
	}

static void
create_clock_tab(GtkWidget *tab_vbox)
	{
	GtkWidget	*tabs;
	GtkWidget	*table, *vbox, *vbox1, *hbox;

	tabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

	vbox = gkrellm_create_framed_tab(tabs, _("Options"));

	vbox1 = gkrellm_framed_vbox(vbox, _("Calendar"), 4, FALSE, 0, 2);
	gkrellm_check_button(vbox1, &cal_enable_button, cal_enable, FALSE, 2,
				_("Enable"));

	vbox1 = gkrellm_framed_vbox_end(vbox, _("Clock"), 4, FALSE, 0, 2);

	gkrellm_check_button(vbox1, &clock_enable_button, clock_enable, 
			FALSE, 6, _("Enable"));

	gkrellm_check_button_connected(vbox1, &clock_24hr_button, clock_24hr_mode,
			FALSE, FALSE, 0, &cb_clock_24hr, 0, 
			_("Display 24 hour instead of 12 hour time"));

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
	gkrellm_check_button(hbox, &clock_12hr_sec_button, clock_12hr_sec_enable,
			FALSE,20, _("Show seconds instead of am/pm in 12 hour time mode"));

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
	gkrellm_check_button(hbox, &clock_24hr_sec_button, clock_24hr_sec_enable,
			FALSE, 20, _("Show seconds in 24 hour time mode"));

	set_seconds_buttons_sensitivity(clock_24hr_mode);

/* -- Setup tab */
	vbox = gkrellm_create_framed_tab(tabs, _("Setup"));

	vbox1 = gkrellm_framed_vbox(vbox, _("Clock Chime commands"),
				4, FALSE, 0, 2);
	table = gkrellm_launcher_table_new(vbox1, 2);

	gkrellm_config_launcher(table, 0, &hour_chime_entry, NULL,
			_("Hour"), NULL);
	gtk_entry_set_text(GTK_ENTRY(hour_chime_entry), hour_chime_command);

	gkrellm_config_launcher(table, 1, &quarter_chime_entry, NULL,
			_("Quarter hour"), NULL);
	gtk_entry_set_text(GTK_ENTRY(quarter_chime_entry), quarter_chime_command);

	gkrellm_check_button(vbox, &loop_chime_button, loop_chime_enable,
			FALSE, 6, _("Loop hour chime command"));


	vbox = gkrellm_framed_vbox_end(vbox, _("Launch Commands"), 4, FALSE, 0, 2);
	table = gkrellm_launcher_table_new(vbox, 2);

	gkrellm_config_launcher(table, 0,  &cal_launch_entry,
				&cal_tooltip_entry, _("Calendar"), &cal_launch);
	gkrellm_config_launcher(table, 1,  &clock_launch_entry,
				&clock_tooltip_entry, _("Clock"), &clock_launch);
	}


static Monitor	monitor_clock =
	{
	N_("Clock"),		/* Name, for config tab.	*/
	MON_CLOCK,			/* Id,  0 if a plugin		*/
	create_clock,		/* The create function		*/
	update_clock,		/* The update function		*/
	create_clock_tab,	/* The config tab create function	*/
	apply_clock_cal_config,	/* Apply the config function		*/

	save_clock_cal_config,	/* Save user conifg			*/
	load_clock_cal_config,	/* Load user config			*/
	CLOCK_CONFIG_KEYWORD, /* config keyword			*/

	NULL,				/* Undef 2	*/
	NULL,				/* Undef 1	*/
	NULL,				/* Undef 0	*/

	0,					/* insert_before_id - place plugin before this mon */

	NULL,				/* Handle if a plugin, filled in by GKrellM		*/
	NULL				/* path if a plugin, filled in by GKrellM		*/
	};

Monitor *
init_clock_monitor(void)
	{
	clock_enable = TRUE;
    monitor_clock.name = _(monitor_clock.name);
	clock_12hr_sec_enable = TRUE;
	clock_24hr_sec_enable = TRUE;
    loop_chime_enable   = FALSE;
    hour_chime_command = g_strdup("");
    quarter_chime_command = g_strdup("");

	clock_style_id = gkrellm_add_meter_style(&monitor_clock, CLOCK_STYLE_NAME);
	mon_clock = &monitor_clock;
	return &monitor_clock;
	}

static Monitor	monitor_cal =
	{
	N_("Calendar"),	/* Name, for config tab.	*/
	MON_CAL,		/* Id,  0 if a plugin		*/
	create_calendar, /* The create function		*/
	NULL,			/* The update function		*/
	NULL,			/* The config tab create function	*/
	NULL,			/* Apply the config function		*/

	NULL,			/* Save user conifg			*/
	NULL,			/* Load user config			*/
	NULL,			/* config keyword			*/

	NULL,			/* Undef 2	*/
	NULL,			/* Undef 1	*/
	NULL,			/* Undef 0	*/

	0,				/* insert_before_id - place plugin before this mon */

	NULL,			/* Handle if a plugin, filled in by GKrellM		*/
	NULL			/* path if a plugin, filled in by GKrellM		*/
	};

Monitor *
init_cal_monitor(void)
	{
	cal_enable = TRUE;
    monitor_cal.name = _(monitor_cal.name);

	cal_style_id = gkrellm_add_meter_style(&monitor_cal, CAL_STYLE_NAME);
	mon_cal = &monitor_cal;
	return &monitor_cal;
	}

